package ij2x.plugin.color;

import java.awt.Color;
import java.awt.Point;
import java.awt.color.ColorSpace;
import java.awt.image.ColorModel;
import java.awt.image.MemoryImageSource;
import java.awt.color.ICC_ColorSpace;
import java.awt.color.ICC_Profile;

/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 *
 * <p>Copyright: Copyright (c) 2011</p>
 *
 * <p>Company: </p>
 *
 * @author not attributable
 * @version 1.0
 */
public abstract class AbstractColorWheel extends MemoryImageSource {
    protected int[] pixels;
    protected int w, h;
    protected ColorSpace colorSpace;
    protected int radialIndex = 1;
    protected int angularIndex = 0;
    protected int verticalIndex = 2;
    protected boolean isPixelsValid = false;
    protected float verticalValue = 1f;
    protected boolean isLookupValid = false;

    public AbstractColorWheel(ColorSpace sys, int w, int h) {
       super(w, h, null, 0, w);
       this.colorSpace = sys;
       pixels = new int[w * h];
       this.w = w;
       this.h = h;
       setAnimated(true);

       newPixels(pixels, ColorModel.getRGBdefault(), 0, w);
   }

    public AbstractColorWheel(ColorSpace sys, int[] pixels, int w, int h) {
        super(w, h, pixels, 0, w);
        this.colorSpace = sys;
        this.w = w;
        this.h = h;
        setAnimated(true);

        newPixels(pixels, ColorModel.getRGBdefault(), 0, w);
    }
    public void setRadialComponentIndex(int newValue) {
        radialIndex = newValue;
        isPixelsValid = false;
    }

    public void setAngularComponentIndex(int newValue) {
        angularIndex = newValue;
        isPixelsValid = false;
    }

    public void setVerticalComponentIndex(int newValue) {
        verticalIndex = newValue;
        isPixelsValid = false;
    }

    public void setVerticalValue(float newValue) {
        isPixelsValid = isPixelsValid && verticalValue == newValue;
        verticalValue = newValue;
    }

    public boolean needsGeneration() {
        return !isPixelsValid;
    }

    public void regenerateColorWheel() {
        if (!isPixelsValid) {
            generateColorWheel();
        }
    }

    public int getRadius() {
        return Math.min(w, h) / 2 - 2;
    }

    protected abstract void generateColorWheel();

    public Point getColorLocation(Color c) {
        float[] components =fromColor(colorSpace, c);
        return getColorLocation(components);
    }


    public abstract Point getColorLocation(float[] components);

    public abstract float[] getColorAt(int x, int y);
    /** Returns the color components in the specified color space from a {@code Color} object.
     */
    public static float[] fromColor(ColorSpace colorSpace, Color c) {
        if (isEqual(c.getColorSpace(), colorSpace)) {
            float[] components = c.getComponents(null);
            return components;
        } else {
            return c.getComponents(colorSpace, null);
        }
    }
    /** Returns true, if the two color spaces are equal. */
    public static boolean isEqual(ColorSpace a, ColorSpace b) {
        if ((a instanceof ICC_ColorSpace) && (b instanceof ICC_ColorSpace)) {
            ICC_ColorSpace aicc = (ICC_ColorSpace) a;
            ICC_ColorSpace bicc = (ICC_ColorSpace) b;
            ICC_Profile ap = aicc.getProfile();
            ICC_Profile bp = bicc.getProfile();
            return ap.equals(bp);
        } else {
            return a.equals(b);
        }
    }
    /** Returns the color components in the specified color space from an
     * rgb value.
     */
    public static float[] fromRGB(ColorSpace colorSpace, int rgb) {
        return fromRGB(colorSpace, (rgb >>> 16) & 0xff, (rgb >>> 8) & 0xff, rgb & 0xff);
    }

    /** Returns the color components in the specified color space from
     * RGB values.
     */
    public static float[] fromRGB(ColorSpace colorSpace, int r, int g, int b) {
        return colorSpace.fromRGB(new float[]{r / 255f, g / 255f, b / 255f});
    }

    /** Returns an rgb value from color components in the specified color space.
     */
    public static int toRGB(ColorSpace colorSpace, float... components) {
        float[] rgb = colorSpace.toRGB(components);

        // If the color is not displayable in RGB, we return transparent black.
        if (rgb[0]<0f||rgb[1]<0f||rgb[2]<0f||rgb[0]>1f||rgb[1]>1f||rgb[2]>1f) {
            return 0;
        }
        return 0xff000000 | ((int) (rgb[0] * 255f) << 16) | ((int) (rgb[1] * 255f) << 8) | (int) (rgb[2] * 255f);
    }



}
